/*********************************************
 * ooscc firmware version 0.0.0.0.0.0.1 alpha beta
 * Coded by Colin Raffel
 * Based on code by Guido Schroder 
 * (http://tuxgraphics.org/electronics/200606/article06061.shtml)
 * Copyright: GPL V2
 *
 * ooscc is the open open sound control controller
 * It sends open sound control messages over ethernet
 * This firmware is for a 32 knob (or ADC) version of the ooscc
 * It is preliminary and should be used at your own risk
 * For more information: http://www.experimentalistsanonymous.com/ooscc	
 *
 *********************************************/
#include <avr/io.h>
// This file defines ip/arp/udp functions
#include "ip_arp_udp.h"
// This is the driver for the enc28j60 ethernet interfacing chip
#include "enc28j60.h"
// This is a little function that Guido made to make coding easier
#include "timeout.h"
// This ensures that cbi and sbi macros exist regardless of compiler
#include "avr_compat.h"
// This is part of the driver for the enc28j60
#include "net.h"
#include "stdlib.h"

// The next two lines MUST be modified.  The mac address and ip address 
// must be unique within your network, so come up with something interesting.
static uint8_t mymac[6] = {0x13,0x33,0x37,0x71,0x77,0x20}; 
static uint8_t myip[4] = {192,168,1,10};
// Listening port for UDP commands, 1200 by default
static uint16_t myport =1200;
// When the ooscc gets an init message, it pairs with that MAC addres
// and IP.  That's what these arrays are for (you don't need to
// modify these).
uint8_t buddymac[6] = {0x00,0x00,0x00,0x00,0x00,0x00};
uint8_t buddyip[4] = {0,0,0,0};

// Define the buffer size for the enc28j60
#define BUFFER_SIZE 250
static uint8_t buf[BUFFER_SIZE+1];
// Buddybuff is a variable that saves the buffer when a packet is received
// (I don't know how to fill the buffer properly otherwise)
static uint8_t buddybuf[BUFFER_SIZE+1];

int main(void){

		// BEGIN VARIABLE DECLARATIONS

        // Packet length variable
        uint16_t plen;
		// i and c are generic counting variables
        uint8_t i=0;
        uint8_t c=0;
		// Bound keeps track of whether or not the ooscc is "bound" to
		// some IP/MAC address.  When an init message is received
		// this becomes "1".  Certain functions only execute
		// after the ooscc is bound (for example, trying to send
		// messsages!)
		uint8_t bound=0;
		// Payload length variable
        uint8_t payloadlen=0;
		// ADC values get averaged to create a noiseless and accurate
		// result.  adc_stream is the stream of ADC values for averaging.
		// Four values for each of 32 pots
		uint16_t adc_stream[32][4];
		// adc_address is the current spot in the stream array we are
		// getting a value for.
		uint8_t adc_address=0;
		// adc_avg is the average of the last four results for a given
		// adc input.  It is compared with adc_old
		uint16_t adc_avg=0;
		// The last adc value sent (compared with the current adc average)
		uint16_t adc_old[32]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
		// Current pot we are checking
		uint8_t adc_count=0;
		// All OSC messages are arrays which are formed in compliance
		// with the osc specification http://opensoundcontrol.org/spec-1_0
		// This is a very kludgy method of doing this, a function that
		// converted a string to an OSC string would be much more 
		// elegant, but if it's not broken don't fix it (for now!).
		// initmess is the message sent when the ooscc receives
		// /init, which tells the sender of the /init message 
		// what ip address it is bound to.
		// Format: /boundto (ip address as 4 ints)
		char initmess[36]={'/','b','o','u','n','d','t','o',0,0,0,0,',','i','i','i','i',0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
		// adcmess is sent whenever an adc value changes.
		// Format: /adc number value
		char adcmess[20]={'/','a','d','c',0,0,0,0,',','i','i',0,0,0,0,0,0,0,0};
        
		// END VARIABLE DECLARATIONS
		
		// BEGIN INITIATION SEQUENCE
		// (5, 4, 3, 2.....)
		// (most comments here are by Guido)
		
        // set the clock speed to 8MHz
        // set the clock prescaler. First write CLKPCE to enable setting of clock the
        // next four instructions.
        CLKPR=(1<<CLKPCE);
        CLKPR=0; // 8 MHZ
        
        /* enable PB0, reset as output */
        DDRB|= (1<<DDB0);

        /* set output to gnd, reset the ethernet chip */
        PORTB &= ~(1<<PB0);
        delay_ms(20);
        /* set output to Vcc, reset inactive */
        PORTB|= (1<<PB0);
        delay_ms(100);
		
		//PD3-7 are the pot selection bits
		//(fed into the 4051's)
		//5 pins = 2^5 = 32 analog inputs 
		DDRD |= (1<<DDD3|1<<DDD4|1<<DDD5|1<<DDD6|1<<DDD7);
		
		//select the 0 pot with the 4051
		PORTD = ((PORTD & 0b00000111) | adc_count<<3);
        
        // LED
        /* enable PB1, LED as output */
        DDRB|= (1<<DDB1);

        /* set output to Vcc, LED off */
        PORTB|= (1<<PB1);
        
        //Enable pullup on pin D0
        PORTD|= (1<<PD0);
		
		//Initialize ADC
		//enable ADC by setting the AD enable bit
		ADCSRA |= (1<<ADEN);
		//Single sample conversion by clearing ADAutoTriggerEnable
		ADCSRA &= ~(1<<ADATE);
		//Select divide clock by 64
		ADCSRA = ((ADCSRA & 0b11111000) | 0b00000110);
		//Make AVCC as Vref
		ADMUX = ((ADMUX & 0b00111111) | 0b01000000);
		//Left adjust ADC Result
		ADMUX &= ~(1<<ADLAR);
		//Select ADC0 for conversion
		ADMUX &= 0b11100000;
		//Make Port C input
		DDRC = 0x00;
		//Make sure the pullups are not enabled
		PORTC = 0x00;
		//Set the adc_old value to the current value to prevent a stream on startup
		//Essentially we are filling the adc_stream and adc_old variables
		//with the value on startup.
		//Maybe this should all go at the "init message received" block
		//Do this 32 times (one for each analog input)
		while(adc_count<32){
			//do a conversion
			ADCSRA |= (1<<ADSC);
			//wait for the conversion to stop
			while(bit_is_set(ADCSRA,ADSC))
			{
			}
			//Place the startup analog value in adc_old
			adc_old[adc_count]=((ADCL) | ((ADCH)<<8));
			//fill the stream with the current value
			while(c<4){
				adc_stream[adc_count][c]=adc_old[adc_count];
				c++;
			}
			//reset c variable for next count
			c=0;
			//increment adc_count variable
			adc_count++;
			//select the next analog input
			PORTD = ((PORTD & 0b00000111) | adc_count<<3);
			ADCSRA |= (1<<ADSC);
		}
		//reset adc_count variable
		adc_count=0;

        /*initialize enc28j60*/
        enc28j60Init(mymac);
        delay_ms(20);
		
        /* Magjack leds configuration, see enc28j60 datasheet, page 11 */
        // LEDB=yellow LEDA=green
        //
        // 0x476 is PHLCON LEDA=links status, LEDB=receive/transmit
        // enc28j60PhyWrite(PHLCON,0b0000 0100 0111 01 10);
        enc28j60PhyWrite(PHLCON,0x476);
        delay_ms(20);
        
        /* set output to GND, red LED on */
        PORTB &= ~(1<<PB1);
        i=1;
        
        //init the ethernet/ip layer:
        init_ip_arp_udp(mymac,myip);
		
		// END INITIATION SEQUENCE

		// BEGIN MAIN PROGRAM LOOP

        while(1){
        		//reset our count variable
        		c=0;
                // get the next new packet:
                plen = enc28j60PacketReceive(BUFFER_SIZE, buf);

                /*plen will be unequal to zero if there is a valid 
                 * packet (without crc error) */
                if(plen==0){
					//if we are bound to a computer/ip/mac address
					if(bound==1){
						//if an a to d conversion has completed
						//(one is always in progress)
						if(bit_is_clear(ADCSRA,ADSC)){
							//set the adc stream value for the count (current pot)
							adc_stream[adc_count][adc_address] = ((ADCL) | ((ADCH)<<8));
							//sum up all adc_stream values for the current pot
							while(c<4){
								adc_avg=adc_avg+adc_stream[adc_count][c];
								c++;
								}
							//shift left 2 = divide by 4 = find the average
							adc_avg=(adc_avg>>2);
							//if the stream average is two away from the old adc value
							if((((adc_avg-1) > adc_old[adc_count]) || ((adc_avg+1) < adc_old[adc_count])) && (adc_avg != 0)){
								//average the old and new adc values
								adc_old[adc_count] = (adc_avg+adc_old[adc_count])>>1;
								//set the first int in adcmess to the pot being converted
								adcmess[15]=adc_count;
								//set the second int to the adc value
								adcmess[18] = (uint8_t)(adc_old[adc_count]>>8);
								adcmess[19] = (uint8_t)adc_old[adc_count];
								//send the UDP message
								//(since bound is 1, we can make udp to buddy)
								make_udp_to_buddy(buddybuf,adcmess,20,myport,buddymac,buddyip);
							}
							//set our average to zero so that it sums correctly next time
							adc_avg=0;
							//increment adc count = check the next pot
							adc_count++;
							//switch the 4051 by outputting adc_count on PD3,4,5,6,7
							PORTD = ((PORTD & 0b00000111) | adc_count<<3);
							//if we have checked all of the pots, increment adc_address and reset adc_count
							if(adc_count>31){
								adc_count=0;
								adc_address++;
								//if we have gone through the whole ring buffer, return to place 0
								if(adc_address>3){
									adc_address=0;
								}
							}
							//start an ADC conversion
							ADCSRA |= (1<<ADSC);
						}
					}
					continue;
                }
                // arp is broadcast if unknown but a host may also
                // verify the mac address by sending it to 
                // a unicast address.
                if(eth_type_is_arp_and_my_ip(buf,plen)){
                        make_arp_answer_from_request(buf,plen);
                        continue;
                }
                // check if ip packets (icmp or udp) are for us:
                if(eth_type_is_ip_and_my_ip(buf,plen)==0){
                        continue;
                }
                        
                if(buf[IP_PROTO_P]==IP_PROTO_ICMP_V && buf[ICMP_TYPE_P]==ICMP_TYPE_ECHOREQUEST_V){
                        // a ping packet, let's send pong
                        make_echo_reply_from_request(buf,plen);
                        continue;
                }
				
				//If we got here then there is a message for us to deal with
                if (buf[IP_PROTO_P]==IP_PROTO_UDP_V){
                        //determine payload length
						payloadlen=buf[UDP_LEN_L_P]-UDP_HEADER_LEN;
						//if it's an init message
						//(if the first five characters are /init and the length is 12)
						if (buf[UDP_DATA_P]=='/' && buf[UDP_DATA_P+1]=='i' && buf[UDP_DATA_P+2]=='n' && buf[UDP_DATA_P+3]=='i' && buf[UDP_DATA_P+4]=='t' && payloadlen==12){
							while(c<6){
							    //copy the mac and ip ddress the message came from
								//to the buddy variables
								buddymac[c]=buf[ETH_SRC_MAC +c];
								buddyip[c]=buf[IP_SRC_P+c];
								c++;
							}
							//create init mess with the buddy IP
							c=0;
							while(c<4){
								initmess[23+4*c]=buddyip[c];
								c++;
							}
							c=0;
							//send init message
							make_udp_to_buddy(buf,initmess,36,myport,buddymac,buddyip);
							//make buddy buffer
							while(c<BUFFER_SIZE+1){
								buddybuf[c]=buf[c];
								c++;
							}
							//we are bound now so make bound = 1
							bound=1;
                        }
						//The following is handling for a /adc message (send specific adc value(s))
						//if it's an adc request without a specific pot asked for
						//(this means send all values)
						if (buf[UDP_DATA_P]=='/' && buf[UDP_DATA_P+1]=='a' && buf[UDP_DATA_P+2]=='d' && buf[UDP_DATA_P+3]=='c' && payloadlen==12 && bound==1){
							//now send adc values for pots 0 to 31
							c=0;
							while(c<32){
								//set the first int of the adc message to the pot we are sending
								adcmess[15]=c;
								//set the second int to the adc value
								adcmess[18] = (uint8_t)(adc_old[c]>>8);
								adcmess[19] = (uint8_t)adc_old[c];
								//send the UDP message
								make_udp_to_buddy(buddybuf,adcmess,20,myport,buddymac,buddyip);
								c++;
							}
						}
						//If value(s) are specified, send the values asked for
						if (buf[UDP_DATA_P]=='/' && buf[UDP_DATA_P+1]=='a' && buf[UDP_DATA_P+2]=='d' && buf[UDP_DATA_P+3]=='c' && payloadlen>12 && bound==1){
							//find the first pot asked for.  
							//One message is sent per value asked for
							i=15+4*((payloadlen-8)/20);
							while(i<payloadlen){
								//send adc value for the pot asked for
								c=buf[UDP_DATA_P+i];
								adcmess[15]=c;
								//set the second int to the adc value
								adcmess[18] = (uint8_t)(adc_old[c]>>8);
								adcmess[19] = (uint8_t)adc_old[c];
								//send the UDP message
								make_udp_to_buddy(buddybuf,adcmess,20,myport,buddymac,buddyip);
								i=i+4;
							}
							i=0;
						}
					}
                }
        
        return (0);
		
		// END MAIN PROGRAM LOOP
}
